from medl.common import SongData, BaseLogger

__all__ = ["LyricsFetcherLogger"]


class LyricsFetcherLogger:
    def __init__(self, logger: BaseLogger) -> None:
        self._logger = logger

    def start_fetching(self, track_data: SongData) -> None:
        self._logger.info(f"{self._log_prefix(track_data)} Fetching lyrics")

    def fetched(self, track_data: SongData) -> None:
        self._logger.debug(f"{self._log_prefix(track_data)} Fetched lyrics")

    def no_result(self, track_data: SongData) -> None:
        self._logger.warning(f"{self._log_prefix(track_data)} Lyrics not found")

    def _log_prefix(self, track_data: SongData) -> str:
        return f"[{track_data.name} - {', '.join(track_data.artists)}]"



from typing import List, Protocol

from medl.common import SongData

__all__ = ["BaseMusicToolbox"]


class BaseMusicToolbox(Protocol):
    def search_and_download(self, queries: List[str]) -> None:
        ...

    def search(
        self,
        query: str,
        limit: int,
        sorted: bool = False,
    ) -> List[SongData]:
        ...

    def fetch(self, query: str) -> List[SongData]:
        ...

    def download(self, tracks: List[SongData]) -> None:
        ...



from typing import Protocol
from medl.common import SongData

__all__ = ["BaseMetadataFetcherLogger"]


class BaseMetadataFetcherLogger(Protocol):
    def start_fetching(self, query: str) -> None:
        ...

    def fetched_track(self, track_data: SongData) -> None:
        ...

    def fetched_playlist(self, playlist_name: str, playlist_length: int) -> None:
        ...

    def no_result(self, query: str) -> None:
        ...



from yt_dlp import YoutubeDL
from medl.main.data import DownloadInfo
from medl.common import MedlException, TEMP_MUSIC_PATH, BaseOptionsManager

__all__ = ["Downloader"]


class Downloader:
    def __init__(self, options_manager: BaseOptionsManager) -> None:
        self._options = options_manager.get_options()

        if self._options.extension == "m4a":
            ytdl_format = "bestaudio[ext=m4a]/bestaudio/best"
        elif self._options.extension == "opus":
            ytdl_format = "bestaudio[ext=webm]/bestaudio/best"
        else:
            ytdl_format = "bestaudio"

        yt_dlp_options = {
            "format": ytdl_format,
            "quiet": True,
            "no_warnings": True,
            "encoding": "UTF-8",
            # "cookiefile": self.cookie_file,
            "outtmpl": f"{TEMP_MUSIC_PATH}/%(id)s.%(ext)s",
            "retries": 5,
        }

        self._youtube_dl = YoutubeDL(yt_dlp_options)

    def download(self, url: str) -> DownloadInfo:
        info = self._youtube_dl.extract_info(url, download=True)

        if info is None:
            raise MedlException("Couldn't download music from URL %s", url)

        return DownloadInfo(
            id=info["id"],
            ext=info["ext"],
            bitrate=info["abr"],
            download_path=f"{TEMP_MUSIC_PATH}/{info['id']}.{info['ext']}",
        )



from medl.common.data import MedlOptions

__all__ = ["OptionsManager"]


class OptionsManager:
    def __init__(self) -> None:
        self._options: MedlOptions = MedlOptions()

    def set_options(self, options: MedlOptions) -> None:
        self._options = options

    def get_options(self) -> MedlOptions:
        return self._options



from .log_formatter import LogFormatter
from .logger import Logger
from .options_manager import OptionsManager
from .initializer import Initializer

__all__ = ["LogFormatter", "Logger", "OptionsManager", "Initializer"]



from typing import Optional, Protocol
from medl.common import SongData

__all__ = ["BaseLyricsFetcher"]


class BaseLyricsFetcher(Protocol):
    def fetch(self, track_data: SongData) -> Optional[str]:
        ...



from typing import List, Protocol

from medl.common import SongData

__all__ = ["BaseMetadataFetcher"]


class BaseMetadataFetcher(Protocol):
    def search(
        self,
        query: str,
        limit: int,
        sorted: bool = False,
    ) -> List[SongData]:
        ...

    def fetch(self, query: str) -> List[SongData]:
        ...



from typing import List
from medl.common import SongData, BaseLogger, BaseOptionsManager
from medl.main.interfaces import BasePathFormatter

__all__ = ["PreDownloadHandler"]


class PreDownloadHandler:
    def __init__(
        self,
        logger: BaseLogger,
        options_manager: BaseOptionsManager,
        path_formatter: BasePathFormatter,
    ) -> None:
        self._logger = logger
        self._options = options_manager.get_options()
        self._path_formatter = path_formatter

    def exec(self, tracks: List[SongData]) -> List[SongData]:
        self._logger.debug("Pre-download operation")
        tracks_to_dl = []

        for track in tracks:
            track_path = self._path_formatter.format(track)

            if track_path.exists():
                if self._options.overwrite == "skip":
                    self._logger.info(f"Skipping {track.name} - {', '.join(track.artists)}")
                    continue
                else:
                    self._logger.warning(
                        f"Overwriting {track.name} - {', '.join(track.artists)}"
                    )

            tracks_to_dl.append(track)

        return tracks_to_dl



from .main import BaseMusicToolbox
from .common import BaseOptionsManager, MedlException, SongData, MedlOptions
from .dependency_injection import add_medl


__all__ = [
    "BaseMusicToolbox",
    "BaseOptionsManager",
    "MedlException",
    "SongData",
    "MedlOptions",
    "add_medl",
]



from typing import Optional
from mumee import YTMusicMetadataClient

from medl.common import SongData
from medl.main.interfaces import BaseDownloadUrlFinderLogger

__all__ = ["DownloadUrlFinder"]


class DownloadUrlFinder:
    def __init__(
        self, logger: BaseDownloadUrlFinderLogger, ytm_client: YTMusicMetadataClient
    ) -> None:
        self._logger = logger
        self._ytm_client = ytm_client

    def search(self, track_data: SongData) -> Optional[str]:
        self._logger.start_searching(track_data)

        try:
            result = self._ytm_client.search(
                f"{track_data.name} - {', '.join(track_data.artists)}", 1, True
            )[0]

            self._logger.found(track_data)
            return result.url
        except:
            self._logger.no_result(track_data)
            return None



from .base_track_converter_logger import BaseTrackConverterLogger
from .base_download_url_finder_logger import BaseDownloadUrlFinderLogger
from .base_lyrics_fetcher_logger import BaseLyricsFetcherLogger
from .base_track_downloader_logger import BaseTrackDownloaderLogger
from .base_metadata_embedder_logger import BaseMetadataEmbedderLogger
from .base_metadata_fetcher_logger import BaseMetadataFetcherLogger

__all__ = [
    "BaseTrackConverterLogger",
    "BaseDownloadUrlFinderLogger",
    "BaseLyricsFetcherLogger",
    "BaseTrackDownloaderLogger",
    "BaseMetadataEmbedderLogger",
    "BaseMetadataFetcherLogger",
]



from typing import Optional
from medl.common import SongData, MedlException
from medl.main.data import DownloadInfo
from medl.main.interfaces import (
    BaseTrackDownloaderLogger,
    BaseDownloader,
)

__all__ = ["TrackDownloader"]


class TrackDownloader:
    def __init__(
        self,
        logger: BaseTrackDownloaderLogger,
        downloader: BaseDownloader,
    ) -> None:
        self._logger = logger
        self._downloader = downloader

    def download(self, track_data: SongData, download_url: str) -> Optional[DownloadInfo]:
        self._logger.start_downloading(track_data, download_url)

        try:
            download_info = self._downloader.download(download_url)

            self._logger.downloaded(track_data)
            return download_info
        except MedlException as ex:
            self._logger.error_downloading(track_data, str(ex))
            return None



from typing import Protocol
from pathlib import Path
from medl.common import SongData

__all__ = ["BasePathFormatter"]


class BasePathFormatter(Protocol):
    def format(self, track_data: SongData) -> Path:
        ...



from typing import List
from mumee import (
    SongMetadata,
    BaseMetadataClient,
    BaseMetadataExplorer,
    SearchMetadataCommand,
)

from medl.common import SongData, BaseOptionsManager
from medl.main.interfaces import BaseMetadataFetcherLogger

__all__ = ["MetadataFetcher"]


class MetadataFetcher:
    def __init__(
        self,
        logger: BaseMetadataFetcherLogger,
        options_manager: BaseOptionsManager,
        metadata_client: BaseMetadataClient,
        metadata_explorer: BaseMetadataExplorer,
    ) -> None:
        self._logger = logger
        self._options = options_manager.get_options()
        self._metadata_client = metadata_client
        self._metadata_explorer = metadata_explorer

    def fetch(self, query: str) -> List[SongData]:
        self._logger.start_fetching(query)

        metadata = self._metadata_client.exec(query)

        if metadata is None:
            self._logger.no_result(query)
            return []

        if isinstance(metadata, SongMetadata):  # song
            track_data = SongData.from_mumee(metadata)

            self._logger.fetched_track(track_data)
            return [track_data]
        else:  # playlist
            track_list = [SongData.from_mumee(track) for track in metadata.tracks]

            self._logger.fetched_playlist(metadata.name, len(metadata.tracks))
            return track_list

    def search(
        self,
        query: str,
        limit: int,
        sorted: bool = False,
    ) -> List[SongData]:
        command = SearchMetadataCommand(
            query=query,
            limit_per_client=limit,
            sorted=sorted,
            clients=self._options.metadata_providers,
        )
        metadatas = self._metadata_explorer.exec(command)

        if metadatas is None:
            self._logger.no_result(query)
            return []

        tracks = [SongData.from_mumee(track) for track in metadatas]
        return tracks



from typing import Protocol

__all__ = ["BaseLogger"]


class BaseLogger(Protocol):
    def set_level(self, log_level: int) -> None:
        ...

    def debug(self, msg: str, *args) -> None:
        ...

    def info(self, msg: str, *args) -> None:
        ...

    def warning(self, msg: str, *args) -> None:
        ...

    def error(self, msg: str, *args) -> None:
        ...

    def critical(self, msg: str, *args) -> None:
        ...



from taipan_di import ServiceCollection

from medl.common import add_common
from medl.main import add_main

__all__ = ["add_medl"]


def add_medl(services: ServiceCollection) -> ServiceCollection:
    add_common(services)
    add_main(services)

    return services



from pathlib import Path
from typing import Optional
from yumee import BaseSongFileProvider

from medl.common import SongData
from medl.main.interfaces import BaseMetadataEmbedderLogger

__all__ = ["MetadataEmbedder"]


class MetadataEmbedder:
    def __init__(
        self,
        logger: BaseMetadataEmbedderLogger,
        song_file_provider: BaseSongFileProvider,
    ) -> None:
        self._logger = logger
        self._song_file_provider = song_file_provider

    def embed(self, track_data: SongData, path: Path, lyrics: Optional[str]) -> bool:
        self._logger.start_embedding(track_data)

        track_data.lyrics = lyrics
        metadata = track_data.to_yumee()

        song_file = self._song_file_provider.exec(path)

        if song_file is None:
            self._logger.error_embedding(track_data, str(path))
            return False
        else:
            song_file.embed(metadata)
            self._logger.embedded(track_data)
            return True



from pathlib import Path
from typing import Optional
from medl.common import SongData, MedlException
from medl.main.data import DownloadInfo
from medl.main.interfaces import (
    BaseTrackConverterLogger,
    BasePathFormatter,
    BaseFfmpegConverter,
)

__all__ = ["TrackConverter"]


class TrackConverter:
    def __init__(
        self,
        logger: BaseTrackConverterLogger,
        path_formatter: BasePathFormatter,
        ffmpeg_converter: BaseFfmpegConverter,
    ) -> None:
        self._logger = logger
        self._formatter = path_formatter
        self._ffmpeg_converter = ffmpeg_converter

    def convert(self, track_data: SongData, download_info: DownloadInfo) -> Optional[Path]:
        self._logger.start_converting(track_data)

        tmp_path = Path(download_info.download_path)
        new_path = self._formatter.format(track_data)

        new_path.resolve().parent.mkdir(parents=True, exist_ok=True)

        try:
            self._ffmpeg_converter.convert(tmp_path, new_path)

            self._logger.converted(track_data)
            return new_path
        except MedlException as ex:
            self._logger.error_converting(track_data, str(ex))
            return None
        finally:
            if tmp_path.exists():
                tmp_path.unlink()



from .download_url_finder_logger import DownloadUrlFinderLogger
from .lyrics_fetcher_logger import LyricsFetcherLogger
from .metadata_embedder_logger import MetadataEmbedderLogger
from .track_converter_logger import TrackConverterLogger
from .track_downloader_logger import TrackDownloaderLogger
from .metadata_fetcher_logger import MetadataFetcherLogger

__all__ = [
    "DownloadUrlFinderLogger",
    "LyricsFetcherLogger",
    "MetadataEmbedderLogger",
    "TrackConverterLogger",
    "TrackDownloaderLogger",
    "MetadataFetcherLogger",
]



from .download_url_finder import DownloadUrlFinder
from .lyrics_fetcher import LyricsFetcher
from .metadata_embedder import MetadataEmbedder
from .track_converter import TrackConverter
from .track_downloader import TrackDownloader
from .metadata_fetcher import MetadataFetcher
from .pre_download_handler import PreDownloadHandler
from .post_download_handler import PostDownloadHandler

__all__ = [
    "DownloadUrlFinder",
    "LyricsFetcher",
    "MetadataEmbedder",
    "TrackConverter",
    "TrackDownloader",
    "MetadataFetcher",
    "PreDownloadHandler",
    "PostDownloadHandler",
]



from typing import Protocol
from medl.common import SongData

__all__ = ["BaseTrackConverterLogger"]


class BaseTrackConverterLogger(Protocol):
    def start_converting(self, track_data: SongData) -> None:
        ...

    def converted(self, track_data: SongData) -> None:
        ...

    def error_converting(self, track_data: SongData, error_message: str) -> None:
        ...



from pathlib import Path
import subprocess
from typing import List

from medl.common import MedlException

__all__ = ["FfmpegConverter"]


class FfmpegConverter:
    FFMPEG_FORMATS = {
        "mp3": ["-codec:a", "libmp3lame"],
        "flac": ["-codec:a", "flac", "-sample_fmt", "s16"],
        "ogg": ["-codec:a", "libvorbis"],
        "opus": ["-codec:a", "libopus"],
        "m4a": ["-codec:a", "aac"],
        "wav": ["-codec:a", "pcm_s16le"],
    }

    def __init__(self):
        pass

    def convert(self, input_file: Path, output_file: Path) -> None:
        arguments: List[str] = [
            "-nostdin",
            "-y",
            "-i",
            str(input_file.resolve()),
            "-movflags",
            "+faststart",
            "-v",
            "debug",
            "-progress",
            "-",
            "-nostats",
        ]

        file_format = input_file.suffix[1:]
        output_format = output_file.suffix[1:]

        # Add output format to command
        # -c:a is used if the file is not an matroska container
        # and we want to convert to opus
        # otherwise we use arguments from FFMPEG_FORMATS
        if output_format == "opus" and file_format != "webm":
            arguments.extend(["-c:a", "libopus"])
        elif (output_format == "opus" and file_format == "webm") or (
            output_format == "m4a" and file_format == "m4a"
        ):
            # Copy the audio stream to the output file
            arguments.extend(["-vn", "-c:a", "copy"])
        else:
            arguments.extend(self.FFMPEG_FORMATS[output_format])

        arguments.append(str(output_file.resolve()))

        with subprocess.Popen(
            ["ffmpeg", *arguments],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=False,
            shell=True,
        ) as process:
            proc_out = process.communicate()  # wait for process

            if process.returncode != 0:  # if error
                # join stdout and stderr and decode to utf-8
                message = b"".join([out for out in proc_out if out]).decode("utf-8")
                raise MedlException(message)



from typing import Optional
from lyrics_client import BaseLyricsClient, FetchLyricsCommand
from medl.common import SongData, BaseOptionsManager
from medl.main.interfaces import BaseLyricsFetcherLogger

__all__ = ["LyricsFetcher"]


class LyricsFetcher:
    def __init__(
        self,
        logger: BaseLyricsFetcherLogger,
        options_manager: BaseOptionsManager,
        lyrics_client: BaseLyricsClient,
    ) -> None:
        self._logger = logger
        self._options = options_manager.get_options()
        self._lyrics_client = lyrics_client

    def fetch(self, track_data: SongData) -> Optional[str]:
        self._logger.start_fetching(track_data)

        lyrics_command = FetchLyricsCommand(
            track_data.name,
            ", ".join(track_data.artists),
            self._options.lyrics_providers,
        )
        lyrics_results = self._lyrics_client.exec(lyrics_command)

        if not lyrics_results:
            self._logger.no_result(track_data)
            return None
        else:
            self._logger.fetched(track_data)
            return [r for r in lyrics_results if r.lyrics is not None][0].lyrics



from taipan_di import ServiceCollection

from medl.main.classes import (
    Downloader,
    FfmpegConverter,
    Parallelizer,
    PathFormatter,
)
from medl.main.interfaces import (
    BaseFfmpegConverter,
    BaseDownloader,
    BaseParallelizer,
    BasePathFormatter,
)

__all__ = ["add_tools"]


def add_tools(services: ServiceCollection) -> ServiceCollection:
    services.register(BaseDownloader).as_factory().with_implementation(Downloader)
    services.register(BaseFfmpegConverter).as_factory().with_implementation(FfmpegConverter)
    services.register(BaseParallelizer).as_factory().with_implementation(Parallelizer)
    services.register(BasePathFormatter).as_factory().with_implementation(PathFormatter)

    return services



import logging
import rich

from rich.logging import RichHandler
from rich.theme import Theme
from rich.traceback import install

from medl.common.classes import LogFormatter

__all__ = ["Logger"]


class Logger:
    THEME = Theme(
        {
            "bar.back": "grey23",
            "bar.complete": "rgb(165,66,129)",
            "bar.finished": "rgb(114,156,31)",
            "bar.pulse": "rgb(165,66,129)",
            "general": "green",
            "nonimportant": "rgb(40,100,40)",
            "progress.data.speed": "red",
            "progress.description": "none",
            "progress.download": "green",
            "progress.filesize": "green",
            "progress.filesize.total": "green",
            "progress.percentage": "green",
            "progress.remaining": "rgb(40,100,40)",
            "logging.level.debug": "blue",
            "logging.level.info": "green",
            "logging.level.warning": "yellow",
            "logging.level.error": "red",
            "logging.level.critical": "bold red",
        }
    )

    def __init__(self, formatter: LogFormatter) -> None:
        self._logger = logging.getLogger("medl")
        self._formatter = formatter

        self._console = rich.get_console()
        self._console.push_theme(self.THEME)

        self._handler = None

        self.set_level(logging.DEBUG)

    def set_level(self, log_level: int) -> None:
        if self._logger.level == log_level:
            return None

        self._logger.setLevel(log_level)

        if self._handler is not None:
            self._logger.removeHandler(self._handler)

        self._handler = self._init_handler(log_level)
        self._logger.addHandler(self._handler)

        install(show_locals=False, extra_lines=1, console=self._console)

    def _init_handler(self, log_level: int) -> RichHandler:
        is_debug = log_level == logging.DEBUG

        handler = RichHandler(
            show_time=is_debug,
            log_time_format="[%X]",
            omit_repeated_times=False,
            console=self._console,
            level=log_level,
            markup=True,
            show_path=is_debug,
            show_level=is_debug,
            rich_tracebacks=True,
        )
        handler.setFormatter(self._formatter)

        return handler

    def debug(self, msg: str, *args) -> None:
        self._logger.debug(msg, *args)

    def info(self, msg: str, *args) -> None:
        self._logger.info(msg, *args)

    def warning(self, msg: str, *args) -> None:
        self._logger.warning(msg, *args)

    def error(self, msg: str, *args) -> None:
        self._logger.error(msg, *args)

    def critical(self, msg: str, *args) -> None:
        self._logger.critical(msg, *args)



from typing import List

from medl.common import BaseLogger, SongData, BaseInitializer
from medl.main.interfaces import (
    BaseParallelizer,
    BaseMetadataFetcher,
    BaseDownloadUrlFinder,
    BaseLyricsFetcher,
    BaseMetadataEmbedder,
    BaseTrackConverter,
    BaseTrackDownloader,
    BasePreDownloadHandler,
    BasePostDownloadHandler,
)

__all__ = ["MusicToolbox"]


class MusicToolbox:
    def __init__(
        self,
        logger: BaseLogger,
        parallelizer: BaseParallelizer,
        pre_download_handler: BasePreDownloadHandler,
        metadata_fetcher: BaseMetadataFetcher,
        download_url_finder: BaseDownloadUrlFinder,
        lyrics_fetcher: BaseLyricsFetcher,
        track_downloader: BaseTrackDownloader,
        track_converter: BaseTrackConverter,
        metadata_embedder: BaseMetadataEmbedder,
        post_download_handler: BasePostDownloadHandler,
        initializer: BaseInitializer,
    ) -> None:
        self._logger = logger
        self._parallelizer = parallelizer

        self._pre_download_handler = pre_download_handler
        self._metadata_fetcher = metadata_fetcher
        self._download_url_finder = download_url_finder
        self._lyrics_fetcher = lyrics_fetcher
        self._track_downloader = track_downloader
        self._track_converter = track_converter
        self._metadata_embedder = metadata_embedder
        self._post_download_handler = post_download_handler

        initializer.init()

    def search_and_download(self, queries: List[str]) -> None:
        self._logger.info("Execution started")

        metadata_results: List[List[SongData]] = self._parallelizer.run_in_parallel(
            self.fetch, queries
        )
        tracks = [track for track_list in metadata_results for track in track_list]

        self.download(tracks)

    def search(
        self,
        query: str,
        limit: int,
        sorted: bool = False,
    ) -> List[SongData]:
        results = self._metadata_fetcher.search(query, limit, sorted)
        return results

    def fetch(self, query: str) -> List[SongData]:
        results = self._metadata_fetcher.fetch(query)
        return results

    def download(self, tracks: List[SongData]) -> None:
        tracks = self._pre_download_handler.exec(tracks)

        self._parallelizer.run_in_parallel(self._download_track, tracks)

        self._post_download_handler.exec(tracks)

    def _download_track(self, track_data: SongData) -> None:
        download_url = self._download_url_finder.search(track_data)
        if download_url is None:
            return

        download_info = self._track_downloader.download(track_data, download_url)
        if download_info is None:
            return

        new_path = self._track_converter.convert(track_data, download_info)
        if new_path is None:
            return

        lyrics = self._lyrics_fetcher.fetch(track_data)
        embedded = self._metadata_embedder.embed(track_data, new_path, lyrics)

        self._logger.info(
            f"[{track_data.name} - {', '.join(track_data.artists)}] Completed"
        )



from .base_download_url_finder import BaseDownloadUrlFinder
from .base_lyrics_fetcher import BaseLyricsFetcher
from .base_metadata_embedder import BaseMetadataEmbedder
from .base_track_converter import BaseTrackConverter
from .base_track_downloader import BaseTrackDownloader
from .base_metadata_fetcher import BaseMetadataFetcher
from .base_pre_download_handler import BasePreDownloadHandler
from .base_post_download_handler import BasePostDownloadHandler

__all__ = [
    "BaseDownloadUrlFinder",
    "BaseLyricsFetcher",
    "BaseMetadataEmbedder",
    "BaseTrackConverter",
    "BaseTrackDownloader",
    "BaseMetadataFetcher",
    "BasePreDownloadHandler",
    "BasePostDownloadHandler",
]



from dataclasses import dataclass
from typing import List, Optional
from mumee import SongMetadata as MumeeSongMetadata
from yumee import SongMetadata as YumeeSongMetadata

__all__ = ["SongData"]


@dataclass
class SongData:
    name: str
    artists: List[str]
    artist: str
    genres: List[str]
    disc_number: Optional[int]
    disc_count: Optional[int]
    album_name: Optional[str]
    album_artist: Optional[str]
    duration: int
    year: Optional[int]
    date: Optional[str]
    track_number: Optional[int]
    track_count: Optional[int]
    explicit: Optional[bool]
    cover_url: Optional[str]
    is_song: bool
    id: str
    url: str
    lyrics: Optional[str]

    @classmethod
    def from_mumee(cls, metadata: MumeeSongMetadata) -> "SongData":
        data = SongData(
            name=metadata.name,
            artists=metadata.artists,
            artist=metadata.artist,
            genres=metadata.genres,
            disc_number=metadata.disc_number,
            disc_count=metadata.disc_count,
            album_name=metadata.album_name,
            album_artist=metadata.album_artist,
            duration=metadata.duration,
            year=metadata.year,
            date=metadata.date,
            track_number=metadata.track_number,
            track_count=metadata.track_count,
            explicit=metadata.explicit,
            cover_url=metadata.cover_url,
            is_song=metadata.is_song,
            id=metadata.id,
            url=metadata.url,
            lyrics=None,
        )

        return data

    def to_yumee(self) -> YumeeSongMetadata:
        metadata = YumeeSongMetadata(
            title=self.name,
            artist=self.artist,
            artists=self.artists,
            album_name=self.album_name,
            album_artist=self.album_artist,
            track_number=self.track_number,
            track_count=self.track_count,
            disc_number=self.disc_number,
            disc_count=self.disc_count,
            genres=self.genres,
            date=self.date,
            year=self.year,
            lyrics=self.lyrics,
            cover_url=self.cover_url,
            explicit=self.explicit,
            origin_website=self.url,
        )

        return metadata



from typing import Optional, Protocol
from medl.common import SongData
from medl.main.data import DownloadInfo

__all__ = ["BaseTrackDownloader"]


class BaseTrackDownloader(Protocol):
    def download(self, track_data: SongData, download_url: str) -> Optional[DownloadInfo]:
        ...



from taipan_di import ServiceCollection

from medl.main.classes import (
    DownloadUrlFinder,
    LyricsFetcher,
    MetadataEmbedder,
    TrackConverter,
    TrackDownloader,
    MetadataFetcher,
    PreDownloadHandler,
    PostDownloadHandler,
)
from medl.main.interfaces import (
    BaseDownloadUrlFinder,
    BaseLyricsFetcher,
    BaseMetadataEmbedder,
    BaseTrackConverter,
    BaseTrackDownloader,
    BaseMetadataFetcher,
    BasePreDownloadHandler,
    BasePostDownloadHandler,
)

__all__ = ["add_handlers"]


def add_handlers(services: ServiceCollection) -> ServiceCollection:
    services.register(BaseDownloadUrlFinder).as_factory().with_implementation(
        DownloadUrlFinder
    )
    services.register(BaseLyricsFetcher).as_factory().with_implementation(LyricsFetcher)
    services.register(BaseMetadataEmbedder).as_factory().with_implementation(
        MetadataEmbedder
    )
    services.register(BaseTrackConverter).as_factory().with_implementation(TrackConverter)
    services.register(BaseTrackDownloader).as_factory().with_implementation(TrackDownloader)
    services.register(BaseMetadataFetcher).as_factory().with_implementation(MetadataFetcher)

    services.register(BasePreDownloadHandler).as_factory().with_implementation(
        PreDownloadHandler
    )
    services.register(BasePostDownloadHandler).as_factory().with_implementation(
        PostDownloadHandler
    )

    return services



from pathlib import Path
from typing import Optional, Protocol
from medl.common import SongData
from medl.main.data import DownloadInfo

__all__ = ["BaseTrackConverter"]


class BaseTrackConverter(Protocol):
    def convert(self, track_data: SongData, download_info: DownloadInfo) -> Optional[Path]:
        ...



from typing import List, TypeVar, Callable, Protocol

__all__ = ["BaseParallelizer"]

T = TypeVar("T")
U = TypeVar("U")


class BaseParallelizer(Protocol):
    def run_in_parallel(self, method: Callable[[T], U], arguments: List[T]) -> List[U]:
        ...



from typing import List
from medl.common import SongData, BaseLogger, BaseOptionsManager
from medl.main.interfaces import BasePathFormatter

__all__ = ["PostDownloadHandler"]


class PostDownloadHandler:
    def __init__(
        self,
        logger: BaseLogger,
        options_manager: BaseOptionsManager,
        path_formatter: BasePathFormatter,
    ) -> None:
        self._logger = logger
        self._options = options_manager.get_options()
        self._path_formatter = path_formatter

    def exec(self, tracks: List[SongData]) -> List[SongData]:
        self._logger.debug("Post-download operation")

        if self._options.generate_m3u:
            # generate m3u file
            return tracks

        return tracks



from .tools import (
    BaseDownloader,
    BaseFfmpegConverter,
    BaseParallelizer,
    BasePathFormatter,
)
from .loggers import (
    BaseTrackConverterLogger,
    BaseDownloadUrlFinderLogger,
    BaseLyricsFetcherLogger,
    BaseTrackDownloaderLogger,
    BaseMetadataEmbedderLogger,
    BaseMetadataFetcherLogger,
)
from .handlers import (
    BaseDownloadUrlFinder,
    BaseLyricsFetcher,
    BaseMetadataEmbedder,
    BaseTrackConverter,
    BaseTrackDownloader,
    BaseMetadataFetcher,
    BasePreDownloadHandler,
    BasePostDownloadHandler,
)
from .base_music_toolbox import BaseMusicToolbox

__all__ = [
    "BaseDownloader",
    "BaseFfmpegConverter",
    "BaseParallelizer",
    "BasePathFormatter",
    "BaseTrackConverterLogger",
    "BaseDownloadUrlFinderLogger",
    "BaseLyricsFetcherLogger",
    "BaseTrackDownloaderLogger",
    "BaseMetadataEmbedderLogger",
    "BaseMetadataFetcherLogger",
    "BaseDownloadUrlFinder",
    "BaseLyricsFetcher",
    "BaseMetadataEmbedder",
    "BaseTrackConverter",
    "BaseTrackDownloader",
    "BaseMetadataFetcher",
    "BasePreDownloadHandler",
    "BasePostDownloadHandler",
    "BaseMusicToolbox",
]



from pathlib import Path
from typing import Protocol

__all__ = ["BaseFfmpegConverter"]


class BaseFfmpegConverter(Protocol):
    def convert(self, input_file: Path, output_file: Path) -> None:
        ...



from taipan_di import ServiceCollection

from medl.common import add_common
from medl.main.di import add_main
from medl.main.classes import MusicToolbox
from medl.common.classes.init import FfmpegInitializer, FolderInitializer


def main():
    FolderInitializer().init()
    FfmpegInitializer().init()

    services = ServiceCollection()
    add_common(services)
    add_main(services)

    provider = services.build()
    toolbox = provider.resolve(MusicToolbox)

    queries = [
        "faint - linkin park",
        "https://music.youtube.com/watch?v=ymo9oX83kJI",
        "https://open.spotify.com/track/5BVNXl23xEtFgsKmbKpYfA?si=d738d02b327142c4",
        "https://music.youtube.com/playlist?list=PLNvqJjX5l6qVcR7mriyRcQjTBw4QGdX1s&feature=share",
    ]
    toolbox.search_and_download(queries)


main()



from taipan_di import ServiceCollection

from medl.common.interfaces import BaseLogger, BaseOptionsManager, BaseInitializer
from medl.common.classes import Logger, LogFormatter, OptionsManager, Initializer

__all__ = ["add_common"]


def add_common(services: ServiceCollection) -> ServiceCollection:
    services.register(BaseLogger).as_singleton().with_implementation(Logger)
    services.register(LogFormatter).as_singleton().with_self()

    services.register(BaseOptionsManager).as_singleton().with_implementation(OptionsManager)

    services.register(BaseInitializer).as_singleton().with_implementation(Initializer)

    return services



from medl.common import SongData, BaseLogger

__all__ = ["MetadataEmbedderLogger"]


class MetadataEmbedderLogger:
    def __init__(self, logger: BaseLogger) -> None:
        self._logger = logger

    def start_embedding(self, track_data: SongData) -> None:
        self._logger.info(f"{self._log_prefix(track_data)} Embedding metadata")

    def embedded(self, track_data: SongData) -> None:
        self._logger.debug(f"{self._log_prefix(track_data)} Metadata embedded")
        ...

    def error_embedding(self, track_data: SongData, path: str) -> None:
        self._logger.error(
            f"{self._log_prefix(track_data)} Format isn't supported or couldn't find path {path}"
        )

    def _log_prefix(self, track_data: SongData) -> str:
        return f"[{track_data.name} - {', '.join(track_data.artists)}]"



from typing import Protocol
from medl.common import SongData

__all__ = ["BaseMetadataEmbedderLogger"]


class BaseMetadataEmbedderLogger(Protocol):
    def start_embedding(self, track_data: SongData) -> None:
        ...

    def embedded(self, track_data: SongData) -> None:
        ...

    def error_embedding(self, track_data: SongData, path: str) -> None:
        ...



from medl.common import SongData, BaseLogger

__all__ = ["MetadataFetcherLogger"]


class MetadataFetcherLogger:
    def __init__(self, logger: BaseLogger) -> None:
        self._logger = logger

    def start_fetching(self, query: str) -> None:
        self._logger.info(f"Fetching metadata for {query}")

    def fetched_track(self, track_data: SongData) -> None:
        self._logger.info(
            f"Found metadata for {track_data.name} - {', '.join(track_data.artists)}"
        )

    def fetched_playlist(self, playlist_name: str, playlist_length: int) -> None:
        self._logger.info(f"Found playlist {playlist_name} with {playlist_length} tracks")

    def no_result(self, query: str) -> None:
        self._logger.error(f"Couldn't find metadata for query {query}")



from typing import Protocol
from medl.common import SongData

__all__ = ["BaseTrackDownloaderLogger"]


class BaseTrackDownloaderLogger(Protocol):
    def start_downloading(self, track_data: SongData, download_url: str) -> None:
        ...

    def downloaded(self, track_data: SongData) -> None:
        ...

    def error_downloading(self, track_data: SongData, error_message: str) -> None:
        ...



from typing import Protocol
from medl.common import SongData

__all__ = ["BaseLyricsFetcherLogger"]


class BaseLyricsFetcherLogger(Protocol):
    def start_fetching(self, track_data: SongData) -> None:
        ...

    def fetched(self, track_data: SongData) -> None:
        ...

    def no_result(self, track_data: SongData) -> None:
        ...



from medl.common import SongData, BaseLogger

__all__ = ["TrackDownloaderLogger"]


class TrackDownloaderLogger:
    def __init__(self, logger: BaseLogger) -> None:
        self._logger = logger

    def start_downloading(self, track_data: SongData, download_url: str) -> None:
        self._logger.info(
            f"{self._log_prefix(track_data)} Downloading with url {download_url}"
        )

    def downloaded(self, track_data: SongData) -> None:
        self._logger.debug(f"{self._log_prefix(track_data)} Downloaded")

    def error_downloading(self, track_data: SongData, error_message: str) -> None:
        self._logger.error(f"{self._log_prefix(track_data)} Download failed")

    def _log_prefix(self, track_data: SongData) -> str:
        return f"[{track_data.name} - {', '.join(track_data.artists)}]"



import asyncio
import sys
from typing import List, TypeVar, Callable
from medl.common import BaseOptionsManager

__all__ = ["Parallelizer"]

T = TypeVar("T")
U = TypeVar("U")


class Parallelizer:
    def __init__(self, options_manager: BaseOptionsManager) -> None:
        options = options_manager.get_options()

        self._loop: asyncio.AbstractEventLoop = (
            asyncio.new_event_loop()
            if sys.platform != "win32"
            else asyncio.ProactorEventLoop()
        )
        asyncio.set_event_loop(self._loop)

        self._semaphore: asyncio.Semaphore = asyncio.Semaphore(options.threads)

    def run_in_parallel(self, method: Callable[[T], U], arguments: List[T]) -> List[U]:
        async def method_with_semaphore(arg: T):
            async with self._semaphore:
                return await self._loop.run_in_executor(None, method, arg)

        tasks = [method_with_semaphore(arg) for arg in arguments]
        results = list(self._loop.run_until_complete(asyncio.gather(*tasks)))

        return results



from taipan_di import ServiceCollection

from medl.main.classes import MusicToolbox
from medl.main.interfaces import BaseMusicToolbox
from medl.main.di import add_tools, add_loggers, add_handlers

from lyrics_client import add_lyrics_client
from yumee import add_yumee
from mumee import add_mumee

__all__ = ["add_main"]


def add_main(services: ServiceCollection) -> ServiceCollection:
    add_mumee(services)
    add_yumee(services)
    add_lyrics_client(services)

    add_tools(services)
    add_loggers(services)
    add_handlers(services)

    services.register(BaseMusicToolbox).as_factory().with_implementation(MusicToolbox)

    return services



from __future__ import annotations
from dataclasses import dataclass, field
from typing import List

from mumee import MetadataClientEnum
from medl.common.consts import MEDL_PATH

__all__ = ["MedlOptions"]


@dataclass
class MedlOptions:
    song_folder: str = f"{str(MEDL_PATH)}/{'{date}'}"
    file_format: str = "{title} - {artists}"
    extension: str = "mp3"
    lyrics_providers: List[str] = field(default_factory=lambda: ["genius"])
    metadata_providers: List[MetadataClientEnum] = field(
        default_factory=lambda: [MetadataClientEnum.ALL]
    )
    overwrite: str = "skip"
    generate_m3u: bool = False
    threads: int = 4

    def copy(self) -> MedlOptions:
        return MedlOptions(
            song_folder=self.song_folder,
            file_format=self.file_format,
            extension=self.extension,
            lyrics_providers=self.lyrics_providers,
            metadata_providers=self.metadata_providers,
            overwrite=self.overwrite,
            generate_m3u=self.generate_m3u,
            threads=self.threads,
        )



from typing import Protocol
from medl.common.data import MedlOptions

__all__ = ["BaseOptionsManager"]


class BaseOptionsManager(Protocol):
    def set_options(self, options: MedlOptions) -> None:
        ...

    def get_options(self) -> MedlOptions:
        ...



from .init import FfmpegInitializer, FolderInitializer

__all__ = ["Initializer"]


class Initializer:
    def __init__(self) -> None:
        pass

    def init(self) -> None:
        FfmpegInitializer.init()
        FolderInitializer.init()



from .consts import TEMP_MUSIC_PATH, MEDL_PATH
from .data import MedlOptions, SongData
from .di import add_common
from .exceptions import MedlException
from .interfaces import BaseLogger, BaseOptionsManager, BaseInitializer

__all__ = [
    "TEMP_MUSIC_PATH",
    "MEDL_PATH",
    "MedlOptions",
    "SongData",
    "add_common",
    "MedlException",
    "BaseLogger",
    "BaseOptionsManager",
    "BaseInitializer",
]



from typing import Protocol, List
from medl.common import SongData

__all__ = ["BasePostDownloadHandler"]


class BasePostDownloadHandler(Protocol):
    def exec(self, tracks: List[SongData]) -> List[SongData]:
        ...



from medl.common import SongData, BaseLogger

__all__ = ["TrackConverterLogger"]


class TrackConverterLogger:
    def __init__(self, logger: BaseLogger) -> None:
        self._logger = logger

    def start_converting(self, track_data: SongData) -> None:
        self._logger.info(f"{self._log_prefix(track_data)} Converting track")

    def converted(self, track_data: SongData) -> None:
        self._logger.debug(f"{self._log_prefix(track_data)} Track converted")

    def error_converting(self, track_data: SongData, error_message: str) -> None:
        self._logger.error(f"{self._log_prefix(track_data)} {error_message}")

    def _log_prefix(self, track_data: SongData) -> str:
        return f"[{track_data.name} - {', '.join(track_data.artists)}]"



from medl.common import TEMP_MUSIC_PATH, MEDL_PATH

__all__ = ["FolderInitializer"]


class FolderInitializer:
    @classmethod
    def init(cls):
        if not MEDL_PATH.exists():
            MEDL_PATH.mkdir()
        if not TEMP_MUSIC_PATH.exists():
            TEMP_MUSIC_PATH.mkdir()



from datetime import date, datetime
from pathlib import Path
from medl.common import SongData, BaseOptionsManager

__all__ = ["PathFormatter"]


class PathFormatter:
    def __init__(self, options_manager: BaseOptionsManager) -> None:
        self._options = options_manager.get_options()

    def format(self, track_data: SongData) -> Path:
        format = f"{self._options.song_folder}/{self._options.file_format}.{self._options.extension}"

        path_str = format
        path_str = path_str.replace("{title}", track_data.name)
        path_str = path_str.replace("{artists}", ", ".join(track_data.artists))
        path_str = path_str.replace(
            "{date}", f"{date.today().day}-{date.today().month}-{date.today().year}"
        )
        path_str = path_str.replace(
            "{time}",
            f"{datetime.now().hour}:{datetime.now().minute}:{datetime.now().second}",
        )

        return Path(path_str)



from typing import Protocol
from medl.common import SongData

__all__ = ["BaseDownloadUrlFinderLogger"]


class BaseDownloadUrlFinderLogger(Protocol):
    def start_searching(self, track_data: SongData) -> None:
        ...

    def found(self, track_data: SongData) -> None:
        ...

    def no_result(self, track_data: SongData) -> None:
        ...



from taipan_di import ServiceCollection

from medl.main.classes import (
    DownloadUrlFinderLogger,
    LyricsFetcherLogger,
    MetadataEmbedderLogger,
    TrackConverterLogger,
    TrackDownloaderLogger,
    MetadataFetcherLogger,
)
from medl.main.interfaces import (
    BaseDownloadUrlFinderLogger,
    BaseLyricsFetcherLogger,
    BaseMetadataEmbedderLogger,
    BaseTrackConverterLogger,
    BaseTrackDownloaderLogger,
    BaseMetadataFetcherLogger,
)

__all__ = ["add_loggers"]


def add_loggers(services: ServiceCollection) -> ServiceCollection:
    services.register(BaseDownloadUrlFinderLogger).as_factory().with_implementation(
        DownloadUrlFinderLogger
    )
    services.register(BaseLyricsFetcherLogger).as_factory().with_implementation(
        LyricsFetcherLogger
    )
    services.register(BaseMetadataEmbedderLogger).as_factory().with_implementation(
        MetadataEmbedderLogger
    )
    services.register(BaseTrackConverterLogger).as_factory().with_implementation(
        TrackConverterLogger
    )
    services.register(BaseTrackDownloaderLogger).as_factory().with_implementation(
        TrackDownloaderLogger
    )
    services.register(BaseMetadataFetcherLogger).as_factory().with_implementation(
        MetadataFetcherLogger
    )

    return services



from .base_downloader import BaseDownloader
from .base_ffmpeg_converter import BaseFfmpegConverter
from .base_parallelizer import BaseParallelizer
from .base_path_formatter import BasePathFormatter

__all__ = [
    "BaseDownloader",
    "BaseFfmpegConverter",
    "BaseParallelizer",
    "BasePathFormatter",
]



from .downloader import Downloader
from .ffmpeg_converter import FfmpegConverter
from .parallelizer import Parallelizer
from .path_formatter import PathFormatter

__all__ = [
    "Downloader",
    "FfmpegConverter",
    "Parallelizer",
    "PathFormatter",
]



from typing import Optional, Protocol

from medl.common import SongData

__all__ = ["BaseDownloadUrlFinder"]


class BaseDownloadUrlFinder(Protocol):
    def search(self, track_data: SongData) -> Optional[str]:
        ...



from .base_logger import BaseLogger
from .base_options_manager import BaseOptionsManager
from .base_initializer import BaseInitializer

__all__ = ["BaseLogger", "BaseOptionsManager", "BaseInitializer"]



from medl.common import SongData, BaseLogger

__all__ = ["DownloadUrlFinderLogger"]


class DownloadUrlFinderLogger:
    def __init__(self, logger: BaseLogger) -> None:
        self._logger = logger

    def start_searching(self, track_data: SongData) -> None:
        self._logger.info(f"{self._log_prefix(track_data)} Searching for download URL")

    def found(self, track_data: SongData) -> None:
        self._logger.debug(f"{self._log_prefix(track_data)} Download URL found !")

    def no_result(self, track_data: SongData) -> None:
        self._logger.error(
            f"{self._log_prefix(track_data)} No result for download URL search !"
        )

    def _log_prefix(self, track_data: SongData) -> str:
        return f"[{track_data.name} - {', '.join(track_data.artists)}]"



from pathlib import Path
from typing import Optional, Protocol

from medl.common import SongData

__all__ = ["BaseMetadataEmbedder"]


class BaseMetadataEmbedder(Protocol):
    def embed(self, track_data: SongData, path: Path, lyrics: Optional[str]) -> bool:
        ...



from .tools import PathFormatter, Downloader, FfmpegConverter, Parallelizer
from .loggers import (
    DownloadUrlFinderLogger,
    LyricsFetcherLogger,
    MetadataEmbedderLogger,
    TrackConverterLogger,
    TrackDownloaderLogger,
    MetadataFetcherLogger,
)
from .handlers import (
    DownloadUrlFinder,
    LyricsFetcher,
    MetadataEmbedder,
    TrackConverter,
    TrackDownloader,
    MetadataFetcher,
    PreDownloadHandler,
    PostDownloadHandler,
)

from .music_toolbox import MusicToolbox

__all__ = [
    "PathFormatter",
    "Downloader",
    "FfmpegConverter",
    "Parallelizer",
    "DownloadUrlFinderLogger",
    "LyricsFetcherLogger",
    "MetadataEmbedderLogger",
    "TrackConverterLogger",
    "TrackDownloaderLogger",
    "MetadataFetcherLogger",
    "DownloadUrlFinder",
    "LyricsFetcher",
    "MetadataEmbedder",
    "TrackConverter",
    "TrackDownloader",
    "MetadataFetcher",
    "PreDownloadHandler",
    "PostDownloadHandler",
    "MusicToolbox",
]



import logging
from logging import LogRecord

from rich.markup import escape

__all__ = ["LogFormatter"]


class LogFormatter(logging.Formatter):
    LEVEL_TO_COLOR = {
        logging.DEBUG: "[blue]",
        logging.INFO: "[green]",
        logging.WARNING: "[yellow]",
        logging.ERROR: "[red]",
        logging.CRITICAL: "[bold red]",
    }

    def format(self, record: LogRecord) -> str:
        std_message = super().format(record)

        if record.levelno == logging.DEBUG:
            std_message = f"{record.threadName} - {std_message}"

        escaped_message = escape(std_message)
        colored_message = f"{self.LEVEL_TO_COLOR[record.levelno]}{escaped_message}"

        return colored_message
